import os
import math

# note that I have no idea where to put yet: TIMs must be *8-bit*

pac_contents_folder = 'pac_contents/'
output_folder = 'cd_contents/PAC/'
BLOCK_SIZE = 2048 

pack_scripts_only = True # for when you're on a slow computer and editing scripts

def repack_pac_folder(folder, output_path):
    print("Packing " + folder)

    fileList = []
    for filename in os.listdir(folder):
        f = os.path.join(folder, filename)
        if os.path.isfile(f) and len(filename) > 0:
            fileList.append(filename) # note we assume the filename is hex

    # each file entry is 4 bytes, (plus 4 bytes for the closer)
    # so for every blocksize/4 files we need another block for the TOC
    sizeOfToc = int(math.ceil((len(fileList) * 4.0 + 4) / BLOCK_SIZE * 1.0))

    tableOfContentsBytes = b""
    currentBlock = sizeOfToc 
    for filename in fileList:
        fullpath = folder + filename

        filename = filename.split("_")[1] # files are in format ####_000000FF; we only want the hex
        filenameWithoutExt = os.path.splitext(filename)[0]
        filenameAsHexInt = int(filenameWithoutExt, 16)

        tableOfContentsBytes += filenameAsHexInt.to_bytes(2, byteorder='little')
        tableOfContentsBytes += currentBlock.to_bytes(2, byteorder='little')

        currentBlock += getBlockSizeOfFile(fullpath)

    # write the terminator
    tableOfContentsBytes += b"\xFF\xFF"
    tableOfContentsBytes += currentBlock.to_bytes(2, byteorder='little')
    tableOfContentsBytes = padBytesToBlockSize(tableOfContentsBytes)

    out_file = open(output_path, "wb")
    out_file.write(tableOfContentsBytes)
    for filename in fileList:
        fullpath = folder + filename
        out_file.write(getFileBytesPaddedToBlockSize(fullpath))
    out_file.close()

def getFileBytesPaddedToBlockSize(path):
    in_file = open(path, "rb")
    data = in_file.read()
    in_file.close()
    return padBytesToBlockSize(data)

def padBytesToBlockSize(data):
    while len(data) % BLOCK_SIZE != 0:
        data += b"\0"
    return data

def getBlockSizeOfFile(path):
    filesize = os.path.getsize(path)
    blocks = int(math.ceil(1.0 * filesize / BLOCK_SIZE))
    return blocks

def repack_all():
    for filename in os.listdir(pac_contents_folder):
        f = os.path.join(pac_contents_folder, filename) + "/"
        if os.path.isdir(f):
            if pack_scripts_only and 'TAK_BIN' not in f:
                continue # skip if not a script file
            
            output_path = output_folder + filename + '.PAC'
            repack_pac_folder(f, output_path)

repack_all()
print("Repacking complete.")
