import os
import math
import sys
import csv
import time
import argparse
import pathlib
import configparser
import difflib

import script_tools

# this is a branch of script_repack.py that creates a new script file containing all the English script,
# then in-place replaces all 95 (not A1) strings with a 67 call to that script, simply to perform one 95 opcode
# as it turns out, doing a 67 for every single line of text is extremely unperformant, so this was abandoned.

def main():
    parser = argparse.ArgumentParser(description='[UNUSED] modifies a Zero4 Champ Doozy-J (PSX) script using a CSV translation file')
    parser.add_argument('-c', '--debug-clean', action='store_true', help='ignores output_path, effectively performing a repack with zero changes. for debugging')
    parser.add_argument('-d', '--debug-diff', action='store_true', help='does not actually output script. instead diffs the bytes of output and output_path to check for changes')
    parser.add_argument('-i', '--config', type=pathlib.Path, help='config file containing overrides and other info')
    parser.add_argument('input_csv', type=pathlib.Path, help='path to translation CSV (format: start pos, end pos, J string, E string)')
    parser.add_argument('input_path', type=pathlib.Path, help='path to original script')
    parser.add_argument('output_path', type=pathlib.Path, help='path to output the final script to (using the same filename as original_script)')
    args = parser.parse_args()
    repack(args)

def repack(args): 
    start_time = time.time()

    input_csv = args.input_csv
    input_path = args.input_path
    output_path = args.output_path

    config = configparser.ConfigParser()
    try:
        if args.config:
            config_path = args.config
        if config_path and config_path.is_file() and config_path.exists(): 
            config.read(config_path)
    except AttributeError:
        pass

    # thanks https://stackoverflow.com/questions/2422177/python-how-can-i-replace-full-width-characters-with-half-width-characters
    HALF2FULL = dict((i, i + 0xFEE0) for i in range(0x21, 0x7F))
    HALF2FULL[0x20] = 0x3000

    class StringReplaceStruct:
        startPos = None
        endPos = None
        stringIndex = None
        newStringBytes = None

    if input_path.is_file() and input_path.exists(): 
        inputFilenameOnly = input_path.name
        orig_file = open(input_path, "rb")
        file_bytes = bytearray(orig_file.read())
        # zeroes are not truncated because sometimes trailing 00s are used in arguments

        input_file = open(input_csv, 'r', encoding='utf-8')
        csv_reader = csv.reader(input_file, delimiter='|', quotechar='{', quoting=csv.QUOTE_MINIMAL, lineterminator='\n')

        stringReplacements = []

        inputFileNumber = int(inputFilenameOnly.split('_')[0]) # assumes input file of title 100_0000000a, for example
        outStringFileIndex = 40960 + inputFileNumber # a000 + number

        for row in csv_reader:
            if args.debug_clean: # ignore all rows if so
                continue

            if len(row) < 5:
                print("WARNING: Invalid CSV row %s" % row)
                continue

            startPos = int(row[0])
            endPos = int(row[1])
            byteName = row[2]
            origString = row[3]
            finalString = row[4]

            # process the final string
            finalString.strip()
            charaReplacementList = [
                ['"', '¨'], # TODO: find a better quote mark
            ]
            for charaTuple in charaReplacementList:
                finalString = finalString.replace(charaTuple[0], charaTuple[1])

            if len(finalString) > 0: # if there is a string to replace
                # make sure the charas in brackets [ ] don't get fullwidth'd, and also remove the brackets
                finalStringIngame = ""
                bracketOpen = False
                for char in finalString:
                    if char == '[':
                        bracketOpen = True
                    elif char == ']':
                        bracketOpen = False
                    else:
                        if bracketOpen == False:
                            finalStringIngame += char.translate(HALF2FULL) # inline fullLen for time
                        else:
                            finalStringIngame += char
                finalStringBytes = finalStringIngame.encode(encoding='shift-jis')

                origStringLength = endPos-startPos
                finalStringSizeDifference = len(finalStringBytes) - origStringLength

                replacement = StringReplaceStruct()
                replacement.startPos = startPos
                replacement.endPos = endPos
                replacement.stringIndex = len(stringReplacements)
                replacement.newStringBytes = finalStringBytes
                stringReplacements.append(replacement)

        # write the string rewrites
        # TODO: this only works for 95 strings and not A1s
        # the string pointers will be shifted later as shown below
        for replacement in stringReplacements:
            availableSpace = replacement.endPos - replacement.startPos
            #break
            if availableSpace < 11:
                print("WARNING: not enough space for replacement at position %d" % replacement.startPos)
            else:
                strIndex = replacement.startPos -3 # including the opcode + address
                while strIndex < replacement.endPos:
                    file_bytes[strIndex] = 0xf0 # no-op byte
                    strIndex += 1

                strIndex = replacement.startPos -3 # including the opcode + address
                file_bytes[strIndex] = 0x67

                destFileBytes = bytearray(outStringFileIndex.to_bytes(4, 'little')) 
                for intIndex in range(0, len(destFileBytes)):
                    file_bytes[strIndex+1+intIndex] = destFileBytes[intIndex]

                indexBytes = bytearray(replacement.stringIndex.to_bytes(4, 'little')) 
                for intIndex in range(0, len(indexBytes)):
                    file_bytes[strIndex+5+intIndex] = indexBytes[intIndex]

        # write the english string file
        outStringFileBytes = bytearray()
        for i in range(0, len(stringReplacements)):
            # placeholder for table of contents
            outStringFileBytes.append(0x02);
            outStringFileBytes.append(0x01);

        for replacement in stringReplacements:
            startAddressInt = len(outStringFileBytes)
            addressBytes = bytearray(startAddressInt.to_bytes(2, 'little')) 
            outStringFileBytes[replacement.stringIndex * 2] = addressBytes[0]
            outStringFileBytes[replacement.stringIndex * 2 + 1] = addressBytes[1]

            outStringFileBytes.append(0x95)

            indexOfEndStringInt = len(outStringFileBytes) + len(replacement.newStringBytes) + 2 # +2 is for the two address bytes
            addressBytes = bytearray(indexOfEndStringInt.to_bytes(2, 'little')) 
            outStringFileBytes.append(addressBytes[0])
            outStringFileBytes.append(addressBytes[1])

            for intIndex in range(0, len(replacement.newStringBytes)):
                outStringFileBytes.append(replacement.newStringBytes[intIndex])

            outStringFileBytes.append(0x68)

        input_file.close()

        # check and print the difference in byte size
        orig_file = open(str(output_path) + "\\" + inputFilenameOnly, "rb")
        original_bytes = bytearray(orig_file.read())
        orig_file.close()

        # trim excess zeroes
        file_bytes = file_bytes.rstrip(b'\0')
        orig_file.close()
        original_bytes = original_bytes.rstrip(b'\0')

        if len(original_bytes) != len(file_bytes):
            print("Filesize change from %d to %d bytes." % (len(original_bytes), len(file_bytes)))
        if script_tools.getBlockSizeOfData(original_bytes) != script_tools.getBlockSizeOfData(file_bytes):
            print("Blocksize change from %d to %d blocks." % (script_tools.getBlockSizeOfData(original_bytes), script_tools.getBlockSizeOfData(file_bytes)))

        # check and print the difference in byte value
        # TODO: this doesn't do smart diffing. sorry. need to use difflib for that
        bytediffs = []
        new_byte_index = 0
        max_index = max(len(file_bytes), len(original_bytes))
        while new_byte_index < max_index:
            if new_byte_index >= len(file_bytes):
                bytediffs.append([new_byte_index, b'\0', original_bytes[new_byte_index]])
            elif new_byte_index >= len(original_bytes):
                bytediffs.append([new_byte_index, file_bytes[new_byte_index], b'\0'])
            elif file_bytes[new_byte_index] != original_bytes[new_byte_index]:
                bytediffs.append([new_byte_index, file_bytes[new_byte_index], original_bytes[new_byte_index]])
            new_byte_index += 1
        if len(bytediffs) > 0:
            print("%d bytes changed." % len(bytediffs))
            diff_index = 0
            while diff_index < 20 and diff_index < len(bytediffs):
                print("Diff at byte %d: change from %02x to %02x" % (bytediffs[diff_index][0], bytediffs[diff_index][2], bytediffs[diff_index][1]))
                diff_index += 1

        # then actually write the change if appropriate
        if args.debug_diff == False:
            file_bytes = script_tools.padBytesToBlockSize(file_bytes)

            output_file = open(str(output_path) + "\\" + inputFilenameOnly, "wb")
            output_file.write(file_bytes)
            output_file.close()

            if len(outStringFileBytes) > 0:
                string_File_bytes = script_tools.padBytesToBlockSize(outStringFileBytes)

                # TODO: sorry this is generated automatically instead of as an arg
                outFileName = str(output_path) + "\\" + ("%d_%08x" % (outStringFileIndex, outStringFileIndex))
                output_file = open(outFileName, "wb")
                output_file.write(string_File_bytes)
                output_file.close()

    print("Script repacking complete in %s seconds." % (time.time() - start_time))

if __name__ == '__main__':
    main()
